!
!
! This is the Fortran module to interface the Recursive Sparse BLAS code from within PSBLAS.
! It has been generated by a M4 script.
!
! supported types       :  real*4 real*8 complex*8 complex*16
! supported operations  :  psb_rsb_scale psb_rsb_getdiag psb_rsb_get_rows_nnz psb_rsb_get_rows_sparse psb_rsb_destroy_sparse_matrix psb_rsb_allocate_sparse_matrix psb_rsb_get_matrix_nnz psb_rsb_infinity_norm psb_rsb_usmm psb_rsb_ussm psb_rsb_usmv psb_rsb_set_elements psb_rsb_set_element psb_rsb_get_element
!
!
      module psb_rsb_mod
        use psb_spmat_type
        implicit none

public

        
        interface psb_rsb_scale
         ! scales each row of A by multiplying it to a value of D
         module procedure  psb_rsb_s_scale &
        &, psb_rsb_d_scale &
        &, psb_rsb_c_scale &
        &, psb_rsb_z_scale &
        & ;
        end interface
        
        interface psb_rsb_getdiag
         ! gets the diagonal of A into array D
         module procedure  psb_rsb_s_getdiag &
        &, psb_rsb_d_getdiag &
        &, psb_rsb_c_getdiag &
        &, psb_rsb_z_getdiag &
        & ;
        end interface
        
        interface psb_rsb_get_rows_nnz
         ! gets the number of nonzeros in the specified rows interval
         module procedure  psb_rsb_s_get_rows_nnz &
        &, psb_rsb_d_get_rows_nnz &
        &, psb_rsb_c_get_rows_nnz &
        &, psb_rsb_z_get_rows_nnz &
        & ;
        end interface
        
        interface psb_rsb_get_rows_sparse
         ! writes in ia,ja,va the row index, column index, and value of nonzeros from row fr to lr
         module procedure  psb_rsb_s_get_rows_sparse &
        &, psb_rsb_d_get_rows_sparse &
        &, psb_rsb_c_get_rows_sparse &
        &, psb_rsb_z_get_rows_sparse &
        & ;
        end interface
        
        interface psb_rsb_destroy_sparse_matrix
         ! frees all allocated resources to the descriptor of matrix A
         module procedure  psb_rsb_s_destroy_sparse_matrix &
        &, psb_rsb_d_destroy_sparse_matrix &
        &, psb_rsb_c_destroy_sparse_matrix &
        &, psb_rsb_z_destroy_sparse_matrix &
        & ;
        end interface
        
        interface psb_rsb_allocate_sparse_matrix
         ! allocates a sparse matrix A
         module procedure  psb_rsb_s_allocate_sparse_matrix &
        &, psb_rsb_d_allocate_sparse_matrix &
        &, psb_rsb_c_allocate_sparse_matrix &
        &, psb_rsb_z_allocate_sparse_matrix &
        & ;
        end interface
        
        interface psb_rsb_get_matrix_nnz
         ! gets the nonzeros count of matrix A
         module procedure  psb_rsb_s_get_matrix_nnz &
        &, psb_rsb_d_get_matrix_nnz &
        &, psb_rsb_c_get_matrix_nnz &
        &, psb_rsb_z_get_matrix_nnz &
        & ;
        end interface
        
        interface psb_rsb_infinity_norm
         ! gets the infinity norm (the maximal sum of rows elements) of matrix A
         module procedure  psb_rsb_s_infinity_norm &
        &, psb_rsb_d_infinity_norm &
        &, psb_rsb_c_infinity_norm &
        &, psb_rsb_z_infinity_norm &
        & ;
        end interface
        
        interface psb_rsb_usmm
         ! multiplication  : c <- beta c + alpha A b
         module procedure  psb_rsb_s_usmm &
        &, psb_rsb_d_usmm &
        &, psb_rsb_c_usmm &
        &, psb_rsb_z_usmm &
        & ;
        end interface
        
        interface psb_rsb_ussm
         ! triangular solve: b <- alpha A^-1 b
         module procedure  psb_rsb_s_ussm &
        &, psb_rsb_d_ussm &
        &, psb_rsb_c_ussm &
        &, psb_rsb_z_ussm &
        & ;
        end interface
        
        interface psb_rsb_usmv
         ! multiplication  : c <- beta c + alpha A b
         module procedure  psb_rsb_s_usmv &
        &, psb_rsb_d_usmv &
        &, psb_rsb_c_usmv &
        &, psb_rsb_z_usmv &
        & ;
        end interface
        
        interface psb_rsb_set_elements
         ! sets the matrix at elements ia,ja (if in the nonzero pattern)
         module procedure  psb_rsb_s_set_elements &
        &, psb_rsb_d_set_elements &
        &, psb_rsb_c_set_elements &
        &, psb_rsb_z_set_elements &
        & ;
        end interface
        
        interface psb_rsb_set_element
         ! sets the matrix at element i,j (if in the nonzero pattern)
         module procedure  psb_rsb_s_set_element &
        &, psb_rsb_d_set_element &
        &, psb_rsb_c_set_element &
        &, psb_rsb_z_set_element &
        & ;
        end interface
        
        interface psb_rsb_get_element
         ! gets the matrix at element i,j (if in the nonzero pattern)
         module procedure  psb_rsb_s_get_element &
        &, psb_rsb_d_get_element &
        &, psb_rsb_c_get_element &
        &, psb_rsb_z_get_element &
        & ;
        end interface
        

        integer, parameter :: psb_rsb_const_success=0
        integer, parameter :: psb_rsb_const_blocking_extra_=8 ! FIXME: temporary
        integer, parameter :: psb_rsb_const_failure=-1 ! value returned by this interface on failure
        integer, parameter :: psb_rsb_const_not_available=-9999 ! value returned by this interface when deactivated

contains

        subroutine psb_rsb_init(info)
          implicit none
          integer::info
          info=psb_rsb_const_success
#ifdef HAVE_RSB_KERNELS
          call rsb_do_fortran_init(info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine

        subroutine psb_rsb_exit(info)
          implicit none
          integer::info
          info=psb_rsb_const_success
#ifdef HAVE_RSB_KERNELS
          call rsb_do_fortran_exit(info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine

        
        
        subroutine psb_rsb_s_scale(a,d,info)
          ! scales each row of A by multiplying it to a value of D
          implicit none
          integer, intent(out) ::info
          type(psb_sspmat_type) :: a ! psblas matrix type (float)
          real*4 :: d(:)
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_scale(a%infoa(psb_const_infoa_rsb_),d,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_d_scale(a,d,info)
          ! scales each row of A by multiplying it to a value of D
          implicit none
          integer, intent(out) ::info
          type(psb_dspmat_type) :: a ! psblas matrix type (double)
          real*8 :: d(:)
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_scale(a%infoa(psb_const_infoa_rsb_),d,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_c_scale(a,d,info)
          ! scales each row of A by multiplying it to a value of D
          implicit none
          integer, intent(out) ::info
          type(psb_cspmat_type) :: a ! psblas matrix type (float complex)
          complex*8 :: d(:)
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_scale(a%infoa(psb_const_infoa_rsb_),d,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_z_scale(a,d,info)
          ! scales each row of A by multiplying it to a value of D
          implicit none
          integer, intent(out) ::info
          type(psb_zspmat_type) :: a ! psblas matrix type (double complex)
          complex*16 :: d(:)
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_scale(a%infoa(psb_const_infoa_rsb_),d,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        
        
        subroutine psb_rsb_s_getdiag(a,d,info)
          ! gets the diagonal of A into array D
          implicit none
          integer, intent(out) ::info
          type(psb_sspmat_type) :: a ! psblas matrix type (float)
          real*4 :: d(:)
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_getdiag(a%infoa(psb_const_infoa_rsb_),d,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_d_getdiag(a,d,info)
          ! gets the diagonal of A into array D
          implicit none
          integer, intent(out) ::info
          type(psb_dspmat_type) :: a ! psblas matrix type (double)
          real*8 :: d(:)
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_getdiag(a%infoa(psb_const_infoa_rsb_),d,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_c_getdiag(a,d,info)
          ! gets the diagonal of A into array D
          implicit none
          integer, intent(out) ::info
          type(psb_cspmat_type) :: a ! psblas matrix type (float complex)
          complex*8 :: d(:)
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_getdiag(a%infoa(psb_const_infoa_rsb_),d,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_z_getdiag(a,d,info)
          ! gets the diagonal of A into array D
          implicit none
          integer, intent(out) ::info
          type(psb_zspmat_type) :: a ! psblas matrix type (double complex)
          complex*16 :: d(:)
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_getdiag(a%infoa(psb_const_infoa_rsb_),d,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        
        
        subroutine psb_rsb_s_get_rows_nnz(a,fr,lr,nnz,info)
          ! gets the number of nonzeros in the specified rows interval
          implicit none
          integer, intent(out) ::info
          type(psb_sspmat_type) :: a ! psblas matrix type (float)
          integer :: fr,lr
          integer :: nnz
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_get_rows_nnz(a%infoa(psb_const_infoa_rsb_),fr,lr,nnz,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_d_get_rows_nnz(a,fr,lr,nnz,info)
          ! gets the number of nonzeros in the specified rows interval
          implicit none
          integer, intent(out) ::info
          type(psb_dspmat_type) :: a ! psblas matrix type (double)
          integer :: fr,lr
          integer :: nnz
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_get_rows_nnz(a%infoa(psb_const_infoa_rsb_),fr,lr,nnz,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_c_get_rows_nnz(a,fr,lr,nnz,info)
          ! gets the number of nonzeros in the specified rows interval
          implicit none
          integer, intent(out) ::info
          type(psb_cspmat_type) :: a ! psblas matrix type (float complex)
          integer :: fr,lr
          integer :: nnz
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_get_rows_nnz(a%infoa(psb_const_infoa_rsb_),fr,lr,nnz,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_z_get_rows_nnz(a,fr,lr,nnz,info)
          ! gets the number of nonzeros in the specified rows interval
          implicit none
          integer, intent(out) ::info
          type(psb_zspmat_type) :: a ! psblas matrix type (double complex)
          integer :: fr,lr
          integer :: nnz
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_get_rows_nnz(a%infoa(psb_const_infoa_rsb_),fr,lr,nnz,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        
        
        subroutine psb_rsb_s_get_rows_sparse(a,val,fr,lr,ia,ja,nnz,nzin,append,has_iren,iren,info)
          ! writes in ia,ja,va the row index, column index, and value of nonzeros from row fr to lr
          implicit none
          integer, intent(out) ::info
          type(psb_sspmat_type) :: a ! psblas matrix type (float)
          real*4 :: val(:)
          integer :: fr,lr
          logical :: append
          logical :: has_iren
          integer :: appendi=0
          integer :: nzin,has_ireni=0,iren(:) ! FIXME, still unused !
          integer :: ia(:),ja(:)
          integer :: nnz
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          if(append)appendi=1
          if(has_iren)has_ireni=1
          call rsb_do_fortran_get_rows_sparse(a%infoa(psb_const_infoa_rsb_),val,fr,lr,ia,ja,nnz,nzin,appendi,has_ireni,iren,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_d_get_rows_sparse(a,val,fr,lr,ia,ja,nnz,nzin,append,has_iren,iren,info)
          ! writes in ia,ja,va the row index, column index, and value of nonzeros from row fr to lr
          implicit none
          integer, intent(out) ::info
          type(psb_dspmat_type) :: a ! psblas matrix type (double)
          real*8 :: val(:)
          integer :: fr,lr
          logical :: append
          logical :: has_iren
          integer :: appendi=0
          integer :: nzin,has_ireni=0,iren(:) ! FIXME, still unused !
          integer :: ia(:),ja(:)
          integer :: nnz
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          if(append)appendi=1
          if(has_iren)has_ireni=1
          call rsb_do_fortran_get_rows_sparse(a%infoa(psb_const_infoa_rsb_),val,fr,lr,ia,ja,nnz,nzin,appendi,has_ireni,iren,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_c_get_rows_sparse(a,val,fr,lr,ia,ja,nnz,nzin,append,has_iren,iren,info)
          ! writes in ia,ja,va the row index, column index, and value of nonzeros from row fr to lr
          implicit none
          integer, intent(out) ::info
          type(psb_cspmat_type) :: a ! psblas matrix type (float complex)
          complex*8 :: val(:)
          integer :: fr,lr
          logical :: append
          logical :: has_iren
          integer :: appendi=0
          integer :: nzin,has_ireni=0,iren(:) ! FIXME, still unused !
          integer :: ia(:),ja(:)
          integer :: nnz
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          if(append)appendi=1
          if(has_iren)has_ireni=1
          call rsb_do_fortran_get_rows_sparse(a%infoa(psb_const_infoa_rsb_),val,fr,lr,ia,ja,nnz,nzin,appendi,has_ireni,iren,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_z_get_rows_sparse(a,val,fr,lr,ia,ja,nnz,nzin,append,has_iren,iren,info)
          ! writes in ia,ja,va the row index, column index, and value of nonzeros from row fr to lr
          implicit none
          integer, intent(out) ::info
          type(psb_zspmat_type) :: a ! psblas matrix type (double complex)
          complex*16 :: val(:)
          integer :: fr,lr
          logical :: append
          logical :: has_iren
          integer :: appendi=0
          integer :: nzin,has_ireni=0,iren(:) ! FIXME, still unused !
          integer :: ia(:),ja(:)
          integer :: nnz
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          if(append)appendi=1
          if(has_iren)has_ireni=1
          call rsb_do_fortran_get_rows_sparse(a%infoa(psb_const_infoa_rsb_),val,fr,lr,ia,ja,nnz,nzin,appendi,has_ireni,iren,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        
        
        subroutine psb_rsb_s_destroy_sparse_matrix(a,info)
          ! frees all allocated resources to the descriptor of matrix A
          implicit none
          integer, intent(out) ::info
          type(psb_sspmat_type) :: a ! psblas matrix type (float)
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_destroy_sparse_matrix(a%infoa(psb_const_infoa_rsb_),info)
          a%infoa(psb_nnz_)=0
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_d_destroy_sparse_matrix(a,info)
          ! frees all allocated resources to the descriptor of matrix A
          implicit none
          integer, intent(out) ::info
          type(psb_dspmat_type) :: a ! psblas matrix type (double)
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_destroy_sparse_matrix(a%infoa(psb_const_infoa_rsb_),info)
          a%infoa(psb_nnz_)=0
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_c_destroy_sparse_matrix(a,info)
          ! frees all allocated resources to the descriptor of matrix A
          implicit none
          integer, intent(out) ::info
          type(psb_cspmat_type) :: a ! psblas matrix type (float complex)
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_destroy_sparse_matrix(a%infoa(psb_const_infoa_rsb_),info)
          a%infoa(psb_nnz_)=0
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_z_destroy_sparse_matrix(a,info)
          ! frees all allocated resources to the descriptor of matrix A
          implicit none
          integer, intent(out) ::info
          type(psb_zspmat_type) :: a ! psblas matrix type (double complex)
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_destroy_sparse_matrix(a%infoa(psb_const_infoa_rsb_),info)
          a%infoa(psb_nnz_)=0
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        
        
        subroutine psb_rsb_s_allocate_sparse_matrix(a,val,ia,ja,nnz,m,k,br,bc,extra,info)
          ! allocates a sparse matrix A
          implicit none
          integer, intent(out) ::info
          type(psb_sspmat_type) :: a ! psblas matrix type (float)
          real*4 :: val(:)
          integer :: ia(:),ja(:)
          integer :: nnz,m,k,br,bc
          character(len=*) :: extra
          integer,parameter :: typecode=83
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_allocate_sparse_matrix(a%infoa(psb_const_infoa_rsb_),val,ia,ja,nnz,m,k,br,bc,extra,typecode,info)
          call psb_rsb_get_matrix_nnz(a,a%infoa(psb_nnz_),info) ! FIXME : this is necessary, because modules dependencies forbid calling get_matrix_nnz from within psb_spmat_type !
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_d_allocate_sparse_matrix(a,val,ia,ja,nnz,m,k,br,bc,extra,info)
          ! allocates a sparse matrix A
          implicit none
          integer, intent(out) ::info
          type(psb_dspmat_type) :: a ! psblas matrix type (double)
          real*8 :: val(:)
          integer :: ia(:),ja(:)
          integer :: nnz,m,k,br,bc
          character(len=*) :: extra
          integer,parameter :: typecode=68
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_allocate_sparse_matrix(a%infoa(psb_const_infoa_rsb_),val,ia,ja,nnz,m,k,br,bc,extra,typecode,info)
          call psb_rsb_get_matrix_nnz(a,a%infoa(psb_nnz_),info) ! FIXME : this is necessary, because modules dependencies forbid calling get_matrix_nnz from within psb_spmat_type !
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_c_allocate_sparse_matrix(a,val,ia,ja,nnz,m,k,br,bc,extra,info)
          ! allocates a sparse matrix A
          implicit none
          integer, intent(out) ::info
          type(psb_cspmat_type) :: a ! psblas matrix type (float complex)
          complex*8 :: val(:)
          integer :: ia(:),ja(:)
          integer :: nnz,m,k,br,bc
          character(len=*) :: extra
          integer,parameter :: typecode=67
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_allocate_sparse_matrix(a%infoa(psb_const_infoa_rsb_),val,ia,ja,nnz,m,k,br,bc,extra,typecode,info)
          call psb_rsb_get_matrix_nnz(a,a%infoa(psb_nnz_),info) ! FIXME : this is necessary, because modules dependencies forbid calling get_matrix_nnz from within psb_spmat_type !
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_z_allocate_sparse_matrix(a,val,ia,ja,nnz,m,k,br,bc,extra,info)
          ! allocates a sparse matrix A
          implicit none
          integer, intent(out) ::info
          type(psb_zspmat_type) :: a ! psblas matrix type (double complex)
          complex*16 :: val(:)
          integer :: ia(:),ja(:)
          integer :: nnz,m,k,br,bc
          character(len=*) :: extra
          integer,parameter :: typecode=90
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_allocate_sparse_matrix(a%infoa(psb_const_infoa_rsb_),val,ia,ja,nnz,m,k,br,bc,extra,typecode,info)
          call psb_rsb_get_matrix_nnz(a,a%infoa(psb_nnz_),info) ! FIXME : this is necessary, because modules dependencies forbid calling get_matrix_nnz from within psb_spmat_type !
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        
        
        subroutine psb_rsb_s_get_matrix_nnz(a,ires,info)
          ! gets the nonzeros count of matrix A
          implicit none
          integer, intent(out) ::info
          type(psb_sspmat_type) :: a ! psblas matrix type (float)
          integer :: ires
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_get_matrix_nnz(a%infoa(psb_const_infoa_rsb_),ires,info)
          ires=a%infoa(psb_nnz_) ! FIXME : this is necessary, because modules dependencies forbid calling get_matrix_nnz from within psb_spmat_type !
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_d_get_matrix_nnz(a,ires,info)
          ! gets the nonzeros count of matrix A
          implicit none
          integer, intent(out) ::info
          type(psb_dspmat_type) :: a ! psblas matrix type (double)
          integer :: ires
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_get_matrix_nnz(a%infoa(psb_const_infoa_rsb_),ires,info)
          ires=a%infoa(psb_nnz_) ! FIXME : this is necessary, because modules dependencies forbid calling get_matrix_nnz from within psb_spmat_type !
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_c_get_matrix_nnz(a,ires,info)
          ! gets the nonzeros count of matrix A
          implicit none
          integer, intent(out) ::info
          type(psb_cspmat_type) :: a ! psblas matrix type (float complex)
          integer :: ires
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_get_matrix_nnz(a%infoa(psb_const_infoa_rsb_),ires,info)
          ires=a%infoa(psb_nnz_) ! FIXME : this is necessary, because modules dependencies forbid calling get_matrix_nnz from within psb_spmat_type !
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_z_get_matrix_nnz(a,ires,info)
          ! gets the nonzeros count of matrix A
          implicit none
          integer, intent(out) ::info
          type(psb_zspmat_type) :: a ! psblas matrix type (double complex)
          integer :: ires
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_get_matrix_nnz(a%infoa(psb_const_infoa_rsb_),ires,info)
          ires=a%infoa(psb_nnz_) ! FIXME : this is necessary, because modules dependencies forbid calling get_matrix_nnz from within psb_spmat_type !
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        
        
        subroutine psb_rsb_s_infinity_norm(a,real_in,trans,info)
          ! gets the infinity norm (the maximal sum of rows elements) of matrix A
          implicit none
          integer, intent(out) ::info
          type(psb_sspmat_type) :: a ! psblas matrix type (float)
          real*4 :: real_in
          real*4 :: in
          character :: trans ! 
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_infinity_norm(a%infoa(psb_const_infoa_rsb_),in,trans,info)
           real_in=in ! FIXME : this is a conversion
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_d_infinity_norm(a,real_in,trans,info)
          ! gets the infinity norm (the maximal sum of rows elements) of matrix A
          implicit none
          integer, intent(out) ::info
          type(psb_dspmat_type) :: a ! psblas matrix type (double)
          real*8 :: real_in
          real*8 :: in
          character :: trans ! 
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_infinity_norm(a%infoa(psb_const_infoa_rsb_),in,trans,info)
           real_in=in ! FIXME : this is a conversion
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_c_infinity_norm(a,real_in,trans,info)
          ! gets the infinity norm (the maximal sum of rows elements) of matrix A
          implicit none
          integer, intent(out) ::info
          type(psb_cspmat_type) :: a ! psblas matrix type (float complex)
          real*4 :: real_in
          complex*8 :: in
          character :: trans ! 
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_infinity_norm(a%infoa(psb_const_infoa_rsb_),in,trans,info)
           real_in=in ! FIXME : this is a conversion
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_z_infinity_norm(a,real_in,trans,info)
          ! gets the infinity norm (the maximal sum of rows elements) of matrix A
          implicit none
          integer, intent(out) ::info
          type(psb_zspmat_type) :: a ! psblas matrix type (double complex)
          real*8 :: real_in
          complex*16 :: in
          character :: trans ! 
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_infinity_norm(a%infoa(psb_const_infoa_rsb_),in,trans,info)
           real_in=in ! FIXME : this is a conversion
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        
        
        subroutine psb_rsb_s_usmm(a,b,c,lb,lc,nrhs,alpha,beta,trans,info)
          ! multiplication  : c <- beta c + alpha A b
          implicit none
          integer, intent(out) ::info
          type(psb_sspmat_type) :: a ! psblas matrix type (float)
          real*4 :: b(:,:),c(:,:)
          integer :: lb,lc,nrhs ! FIXME : new, still unused! (FIXME: nrhs is iwsz)
          real*4 :: alpha,beta
          character :: trans ! 
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_usmm(a%infoa(psb_const_infoa_rsb_),b,c,lb,lc,nrhs,alpha,beta,trans,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_d_usmm(a,b,c,lb,lc,nrhs,alpha,beta,trans,info)
          ! multiplication  : c <- beta c + alpha A b
          implicit none
          integer, intent(out) ::info
          type(psb_dspmat_type) :: a ! psblas matrix type (double)
          real*8 :: b(:,:),c(:,:)
          integer :: lb,lc,nrhs ! FIXME : new, still unused! (FIXME: nrhs is iwsz)
          real*8 :: alpha,beta
          character :: trans ! 
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_usmm(a%infoa(psb_const_infoa_rsb_),b,c,lb,lc,nrhs,alpha,beta,trans,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_c_usmm(a,b,c,lb,lc,nrhs,alpha,beta,trans,info)
          ! multiplication  : c <- beta c + alpha A b
          implicit none
          integer, intent(out) ::info
          type(psb_cspmat_type) :: a ! psblas matrix type (float complex)
          complex*8 :: b(:,:),c(:,:)
          integer :: lb,lc,nrhs ! FIXME : new, still unused! (FIXME: nrhs is iwsz)
          complex*8 :: alpha,beta
          character :: trans ! 
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_usmm(a%infoa(psb_const_infoa_rsb_),b,c,lb,lc,nrhs,alpha,beta,trans,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_z_usmm(a,b,c,lb,lc,nrhs,alpha,beta,trans,info)
          ! multiplication  : c <- beta c + alpha A b
          implicit none
          integer, intent(out) ::info
          type(psb_zspmat_type) :: a ! psblas matrix type (double complex)
          complex*16 :: b(:,:),c(:,:)
          integer :: lb,lc,nrhs ! FIXME : new, still unused! (FIXME: nrhs is iwsz)
          complex*16 :: alpha,beta
          character :: trans ! 
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_usmm(a%infoa(psb_const_infoa_rsb_),b,c,lb,lc,nrhs,alpha,beta,trans,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        
        
        subroutine psb_rsb_s_ussm(a,b,c,lb,lc,nrhs,alpha,trans,info)
          ! triangular solve: b <- alpha A^-1 b
          implicit none
          integer, intent(out) ::info
          type(psb_sspmat_type) :: a ! psblas matrix type (float)
          real*4 :: b(:,:),c(:,:)
          integer :: lb,lc,nrhs ! FIXME : new, still unused! (FIXME: nrhs is iwsz)
          real*4 :: alpha
          character :: trans ! 
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_ussm(a%infoa(psb_const_infoa_rsb_),b,c,lb,lc,nrhs,alpha,trans,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_d_ussm(a,b,c,lb,lc,nrhs,alpha,trans,info)
          ! triangular solve: b <- alpha A^-1 b
          implicit none
          integer, intent(out) ::info
          type(psb_dspmat_type) :: a ! psblas matrix type (double)
          real*8 :: b(:,:),c(:,:)
          integer :: lb,lc,nrhs ! FIXME : new, still unused! (FIXME: nrhs is iwsz)
          real*8 :: alpha
          character :: trans ! 
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_ussm(a%infoa(psb_const_infoa_rsb_),b,c,lb,lc,nrhs,alpha,trans,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_c_ussm(a,b,c,lb,lc,nrhs,alpha,trans,info)
          ! triangular solve: b <- alpha A^-1 b
          implicit none
          integer, intent(out) ::info
          type(psb_cspmat_type) :: a ! psblas matrix type (float complex)
          complex*8 :: b(:,:),c(:,:)
          integer :: lb,lc,nrhs ! FIXME : new, still unused! (FIXME: nrhs is iwsz)
          complex*8 :: alpha
          character :: trans ! 
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_ussm(a%infoa(psb_const_infoa_rsb_),b,c,lb,lc,nrhs,alpha,trans,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_z_ussm(a,b,c,lb,lc,nrhs,alpha,trans,info)
          ! triangular solve: b <- alpha A^-1 b
          implicit none
          integer, intent(out) ::info
          type(psb_zspmat_type) :: a ! psblas matrix type (double complex)
          complex*16 :: b(:,:),c(:,:)
          integer :: lb,lc,nrhs ! FIXME : new, still unused! (FIXME: nrhs is iwsz)
          complex*16 :: alpha
          character :: trans ! 
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_ussm(a%infoa(psb_const_infoa_rsb_),b,c,lb,lc,nrhs,alpha,trans,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        
        
        subroutine psb_rsb_s_usmv(a,b,c,alpha,beta,trans,info)
          ! multiplication  : c <- beta c + alpha A b
          implicit none
          integer, intent(out) ::info
          type(psb_sspmat_type) :: a ! psblas matrix type (float)
          real*4 :: b(:),c(:)
          real*4 :: alpha,beta
          character :: trans ! 
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_usmv(a%infoa(psb_const_infoa_rsb_),b,c,alpha,beta,trans,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_d_usmv(a,b,c,alpha,beta,trans,info)
          ! multiplication  : c <- beta c + alpha A b
          implicit none
          integer, intent(out) ::info
          type(psb_dspmat_type) :: a ! psblas matrix type (double)
          real*8 :: b(:),c(:)
          real*8 :: alpha,beta
          character :: trans ! 
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_usmv(a%infoa(psb_const_infoa_rsb_),b,c,alpha,beta,trans,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_c_usmv(a,b,c,alpha,beta,trans,info)
          ! multiplication  : c <- beta c + alpha A b
          implicit none
          integer, intent(out) ::info
          type(psb_cspmat_type) :: a ! psblas matrix type (float complex)
          complex*8 :: b(:),c(:)
          complex*8 :: alpha,beta
          character :: trans ! 
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_usmv(a%infoa(psb_const_infoa_rsb_),b,c,alpha,beta,trans,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_z_usmv(a,b,c,alpha,beta,trans,info)
          ! multiplication  : c <- beta c + alpha A b
          implicit none
          integer, intent(out) ::info
          type(psb_zspmat_type) :: a ! psblas matrix type (double complex)
          complex*16 :: b(:),c(:)
          complex*16 :: alpha,beta
          character :: trans ! 
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_usmv(a%infoa(psb_const_infoa_rsb_),b,c,alpha,beta,trans,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        
        
        subroutine psb_rsb_s_set_elements(a,val,ia,ja,nnz,imin,imax,jmin,jmax,has_gtl,gtl,info)
          ! sets the matrix at elements ia,ja (if in the nonzero pattern)
          implicit none
          integer, intent(out) ::info
          type(psb_sspmat_type) :: a ! psblas matrix type (float)
          integer :: ia(:),ja(:)
          integer :: imin,imax,jmin,jmax,nnz
          integer :: has_gtli,gtl(:)
          logical :: has_gtl
          real*4 :: val(:)
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_set_elements(a%infoa(psb_const_infoa_rsb_),val,ia,ja,nnz,imin,imax,jmin,jmax,has_gtli,gtl,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_d_set_elements(a,val,ia,ja,nnz,imin,imax,jmin,jmax,has_gtl,gtl,info)
          ! sets the matrix at elements ia,ja (if in the nonzero pattern)
          implicit none
          integer, intent(out) ::info
          type(psb_dspmat_type) :: a ! psblas matrix type (double)
          integer :: ia(:),ja(:)
          integer :: imin,imax,jmin,jmax,nnz
          integer :: has_gtli,gtl(:)
          logical :: has_gtl
          real*8 :: val(:)
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_set_elements(a%infoa(psb_const_infoa_rsb_),val,ia,ja,nnz,imin,imax,jmin,jmax,has_gtli,gtl,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_c_set_elements(a,val,ia,ja,nnz,imin,imax,jmin,jmax,has_gtl,gtl,info)
          ! sets the matrix at elements ia,ja (if in the nonzero pattern)
          implicit none
          integer, intent(out) ::info
          type(psb_cspmat_type) :: a ! psblas matrix type (float complex)
          integer :: ia(:),ja(:)
          integer :: imin,imax,jmin,jmax,nnz
          integer :: has_gtli,gtl(:)
          logical :: has_gtl
          complex*8 :: val(:)
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_set_elements(a%infoa(psb_const_infoa_rsb_),val,ia,ja,nnz,imin,imax,jmin,jmax,has_gtli,gtl,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_z_set_elements(a,val,ia,ja,nnz,imin,imax,jmin,jmax,has_gtl,gtl,info)
          ! sets the matrix at elements ia,ja (if in the nonzero pattern)
          implicit none
          integer, intent(out) ::info
          type(psb_zspmat_type) :: a ! psblas matrix type (double complex)
          integer :: ia(:),ja(:)
          integer :: imin,imax,jmin,jmax,nnz
          integer :: has_gtli,gtl(:)
          logical :: has_gtl
          complex*16 :: val(:)
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_set_elements(a%infoa(psb_const_infoa_rsb_),val,ia,ja,nnz,imin,imax,jmin,jmax,has_gtli,gtl,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        
        
        subroutine psb_rsb_s_set_element(a,v,i,j,info)
          ! sets the matrix at element i,j (if in the nonzero pattern)
          implicit none
          integer, intent(out) ::info
          type(psb_sspmat_type) :: a ! psblas matrix type (float)
          integer :: i,j
          real*4 :: v
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_set_element(a%infoa(psb_const_infoa_rsb_),v,i,j,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_d_set_element(a,v,i,j,info)
          ! sets the matrix at element i,j (if in the nonzero pattern)
          implicit none
          integer, intent(out) ::info
          type(psb_dspmat_type) :: a ! psblas matrix type (double)
          integer :: i,j
          real*8 :: v
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_set_element(a%infoa(psb_const_infoa_rsb_),v,i,j,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_c_set_element(a,v,i,j,info)
          ! sets the matrix at element i,j (if in the nonzero pattern)
          implicit none
          integer, intent(out) ::info
          type(psb_cspmat_type) :: a ! psblas matrix type (float complex)
          integer :: i,j
          complex*8 :: v
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_set_element(a%infoa(psb_const_infoa_rsb_),v,i,j,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_z_set_element(a,v,i,j,info)
          ! sets the matrix at element i,j (if in the nonzero pattern)
          implicit none
          integer, intent(out) ::info
          type(psb_zspmat_type) :: a ! psblas matrix type (double complex)
          integer :: i,j
          complex*16 :: v
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_set_element(a%infoa(psb_const_infoa_rsb_),v,i,j,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        
        
        subroutine psb_rsb_s_get_element(a,v,i,j,info)
          ! gets the matrix at element i,j (if in the nonzero pattern)
          implicit none
          integer, intent(out) ::info
          type(psb_sspmat_type) :: a ! psblas matrix type (float)
          integer :: i,j
          real*4 :: v
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_get_element(a%infoa(psb_const_infoa_rsb_),v,i,j,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_d_get_element(a,v,i,j,info)
          ! gets the matrix at element i,j (if in the nonzero pattern)
          implicit none
          integer, intent(out) ::info
          type(psb_dspmat_type) :: a ! psblas matrix type (double)
          integer :: i,j
          real*8 :: v
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_get_element(a%infoa(psb_const_infoa_rsb_),v,i,j,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_c_get_element(a,v,i,j,info)
          ! gets the matrix at element i,j (if in the nonzero pattern)
          implicit none
          integer, intent(out) ::info
          type(psb_cspmat_type) :: a ! psblas matrix type (float complex)
          integer :: i,j
          complex*8 :: v
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_get_element(a%infoa(psb_const_infoa_rsb_),v,i,j,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        subroutine psb_rsb_z_get_element(a,v,i,j,info)
          ! gets the matrix at element i,j (if in the nonzero pattern)
          implicit none
          integer, intent(out) ::info
          type(psb_zspmat_type) :: a ! psblas matrix type (double complex)
          integer :: i,j
          complex*16 :: v
#ifdef HAVE_RSB_KERNELS
          info=psb_rsb_const_success
          call rsb_do_fortran_get_element(a%infoa(psb_const_infoa_rsb_),v,i,j,info)
          if(info.ne.psb_rsb_const_success)info=psb_rsb_const_failure
#else
          info=psb_rsb_const_not_available
#endif
        end subroutine
        
        
      end module psb_rsb_mod
